/*************************************************************************
***     Authentication, authorization, accounting + firewalling package
***     (c) 1998-2003 Anton Vinokurov <anton@netams.com>
***		(c) 2003 NeTAMS Development Team
***		All rights reserved. See 'Copying' file included in distribution
***		For latest version and more info, visit this project web page
***		located at http://www.netams.com
***		 					
*************************************************************************/
/* $Id: net_units.c,v 1.14 2003/04/25 19:40:58 jura Exp $ */

#include "netams.h"

/////////////////////////////////////////////////////////////////////////
// NetUnit class
NetUnit::NetUnit(netunit_type t) {
	id=newOid(t);
	name=NULL;
	parent=NULL;
	type=t;
	next=NULL;

	if (ProcessorCfg.def) {
		ProcessorCfg.def->ap.SetForUnit(POLICY_ACCT, this);
		ProcessorCfg.def->fp.SetForUnit(POLICY_FW, this);
		}
	
	dsl_root=NULL;
	nlp=0;
	monitor=NULL;

	sys_policy_name=SysPolicy_name[SP_NONE]; 
	sys_policy=SP_NONE; 
	sys_policy_perm=0;
	open=0;
	
	}

NetUnit::~NetUnit() {
	this->unit2trees(0);
        DSList *ptr,*dsl=dsl_root;
        while(dsl) {
		ptr=dsl;
                dsl=dsl->next;
                aFree(ptr);
        }
	aFree(name);
}

void NetUnit::setName(char *n){
	aFree(name);
	name=(char *)aMalloc(strlen(n));
	strcpy(name, n); 
	}

match NetUnit::Check(void *s){
	switch(type) {
		case NETUNIT_HOST: return ((NetUnit_host*)this)->Check((struct ip*)s); break;
		case NETUNIT_CLUSTER: return ((NetUnit_cluster*)this)->Check((struct ip*)s); break;
		case NETUNIT_NET: return ((NetUnit_net*)this)->Check((struct ip*)s); break;
		case NETUNIT_GROUP: return ((NetUnit_group*)this)->Check((struct ip*)s); break;
		case NETUNIT_USER: return ((NetUnit_user*)this)->Check((struct ip*)s); break;
		default: break;						 
		}
	return MATCH_NONE;
	}

void NetUnit::unit2tree(IPTree *iptree,unsigned flag) {
	        switch(type) {
                case NETUNIT_HOST: return ((NetUnit_host*)this)->unit2tree(iptree,flag); break;
                case NETUNIT_CLUSTER: return ((NetUnit_cluster*)this)->unit2tree(iptree,flag); break;
                case NETUNIT_NET: return ((NetUnit_net*)this)->unit2tree(iptree,flag); break;
                case NETUNIT_USER: return ((NetUnit_user*)this)->unit2tree(iptree,flag); break;
                default: break;
                }
}

void NetUnit::unit2trees(unsigned flag) {
	Service *s=NULL;

        while((s=Services.getServiceByName("data-source",s))) 
		if(this->checkDSList(s->instance)) {
			IPTree *iptree=((ServiceDS_cfg*)s->cfg)->IPtree;
                	this->unit2tree(iptree,flag);	
		}
}

int NetUnit::setDSList(unsigned ds_id){
	DSList *dsl;
	for(dsl=dsl_root;dsl!=NULL;dsl=dsl->next)
		if (dsl->id==ds_id) return -1; // we will not add existing DS to list
		
	dsl=(DSList*)aMalloc(sizeof(DSList));
	dsl->prev=NULL;
	dsl->id=ds_id;
	if(dsl_root) dsl_root->prev=dsl;
	dsl->next=dsl_root;
	dsl_root=dsl;
	return 0;
	}

int NetUnit::clearDSList(unsigned ds_id){
	for (DSList *dsl=dsl_root;dsl!=NULL;dsl=dsl->next) 
		if (dsl->id==ds_id) {
			if(dsl==dsl_root) dsl_root=dsl->next;
			else dsl->prev->next=dsl->next;
			if(dsl->next) dsl->next->prev=dsl->prev;
			aFree(dsl);
			break;
			}
		
	return 0;
	}

struct DSList *NetUnit::getDSList(){
	return dsl_root;
	}

int NetUnit::checkDSList(unsigned ds_id){
	if(!dsl_root) return 2;
	for(DSList *dsl=dsl_root;dsl!=NULL;dsl=dsl->next)
                if (dsl->id==ds_id) return 1;
	return 0;
	} 
/////////////////////////////////////////////////////////////////////////
// NetUnitsList class
NetUnitsList::NetUnitsList(){
	root=last=NULL;
	num_units=0;
	tries_lock=tries_lock_failed=0;

	rwlock=(pthread_rwlock_t*)aMalloc(sizeof (pthread_rwlock_t));
	pthread_rwlock_init(rwlock, NULL);
	}

NetUnitsList::~NetUnitsList(){
	pthread_rwlock_destroy(rwlock);
	aFree(rwlock);
	}

void NetUnitsList::Insert(NetUnit *s){
	int err=pthread_rwlock_trywrlock(rwlock); 
	if (err==EBUSY) { tries_lock_failed++; pthread_rwlock_wrlock(rwlock); }

	NetUnit *d;
	for(d=root; d!=NULL; d=d->next)
		if (d==s) { 
			if (err!=EDEADLK) pthread_rwlock_unlock(rwlock);
			return;
			}
	if (root==NULL) root=s;
	else last->next=s;
	last=s; 
	num_units++;
	if (err!=EDEADLK) pthread_rwlock_unlock(rwlock);
	}

void NetUnitsList::Delete(NetUnit *s){
	int err=pthread_rwlock_trywrlock(rwlock); 
	if (err==EBUSY) { tries_lock_failed++; pthread_rwlock_wrlock(rwlock); }

	NetUnit *d, *p;
	for(d=root; d!=NULL; d=d->next)	{
		if (d==s) {
			if (s==root && s==last ) root=last=NULL;
			else if (s==root) root=s->next;
			else if (s==last) { last=p; last->next=NULL; }
			else p->next=s->next;
			num_units--;
			break;
			}
		p=d; 
	 	}

	for(d=root; d!=NULL; d=d->next)	if (d->parent==s) d->parent=NULL;

	if (err!=EDEADLK) pthread_rwlock_unlock(rwlock);
	}

void NetUnitsList::Delete(oid id){
	NetUnit *u;
	u=getUnitById(id);
	if (u) Delete(u);
	}

NetUnit* NetUnitsList::getUnit(char *name){
	tries_lock++;
	int err=pthread_rwlock_tryrdlock(rwlock); 
	if (err==EBUSY) { tries_lock_failed++; pthread_rwlock_rdlock(rwlock); }

	NetUnit *d;
	for(d=root; d!=NULL; d=d->next)	{
		if (!strcmp(d->name, name)) {
			if (err!=EDEADLK) pthread_rwlock_unlock(rwlock);
			return d;
			}
	 	}
	if (err!=EDEADLK) pthread_rwlock_unlock(rwlock);
	return NULL;
	}

NetUnit* NetUnitsList::getUnitById(oid id){
	tries_lock++;
	int err=pthread_rwlock_tryrdlock(rwlock); 
	if (err==EBUSY) { tries_lock_failed++; pthread_rwlock_rdlock(rwlock); }

	NetUnit *d;
	for(d=root; d!=NULL; d=d->next)	{
		if (d->id==id) {
			if (err!=EDEADLK) pthread_rwlock_unlock(rwlock);
			return d;
			}
	 	}
	if (err!=EDEADLK) pthread_rwlock_unlock(rwlock);
	return NULL;
	}

void NetUnitsList::DeletePolicyElsewhere(Policy *p){
	NetUnit *d;
	policy_data *cpd;
	tries_lock++;

	int err=pthread_rwlock_trywrlock(rwlock); 
	if (err==EBUSY) { tries_lock_failed++; pthread_rwlock_wrlock(rwlock); }
	else if (err!=EDEADLK) pthread_rwlock_wrlock(rwlock);

	for (d=root; d!=NULL; d=d->next) {
		for (cpd=d->ap.root; cpd!=NULL; cpd=cpd->next) if (cpd->policy==p) d->ap.Delete(p);
		for (cpd=d->fp.root; cpd!=NULL; cpd=cpd->next) if (cpd->policy==p) d->fp.Delete(p);
		}
	if (err!=EDEADLK) pthread_rwlock_unlock(rwlock);
	}

unsigned NetUnitsList::setMon(void *value, oid id, char *name){
	int res=0;
	int err=pthread_rwlock_tryrdlock(rwlock); 
	if (err==EBUSY) { tries_lock_failed++; pthread_rwlock_rdlock(rwlock); }

	NetUnit *d;
	for(d=root; d!=NULL; d=d->next)	{
		if (id && d->id==id) { d->monitor=value; res=1; }
		if (name && !strcmp(name, d->name)) { d->monitor=value; res=1; }
		if (!id && !name) { d->monitor=value; res=1; }
	 	}

	if (err!=EDEADLK) pthread_rwlock_unlock(rwlock);
	return res;
	}

void NetUnitsList::showMon(FILE *out, unsigned to){ //1==show, 0==cfg)
	int err=pthread_rwlock_tryrdlock(rwlock); 
	if (err==EBUSY) { tries_lock_failed++; pthread_rwlock_rdlock(rwlock); }

	NetUnit *d;
	for(d=root; d!=NULL; d=d->next)	{
		if (d->monitor && d->name) {
			if (to) fprintf(out, "%s(%06X) ", d->name, d->id);
			else fprintf(out, "monitor unit %06X\n", d->id);
			}
	 	}

	if (err!=EDEADLK) pthread_rwlock_unlock(rwlock);
	return;
	}

/////////////////////////////////////////////////////////////////////////
// NetUnit_cluster class
NetUnit_cluster::NetUnit_cluster() : NetUnit(NETUNIT_CLUSTER) {
	root=last=NULL;
	references=0;
	lock=(pthread_mutex_t*)aMalloc(sizeof (pthread_mutex_t));
	pthread_mutex_init(lock, NULL);
	}

NetUnit_cluster::~NetUnit_cluster(){
	pthread_mutex_destroy(lock);
	}

void NetUnit_cluster::Add(struct in_addr ip){
	pthread_mutex_lock(lock);
	NetUnit_host *d, *s;
	for(d=root; d!=NULL; d=d->next)
		if (d->ip.s_addr==ip.s_addr) { 
			pthread_mutex_unlock(lock);
			return;
			}
	s=(NetUnit_host *)aMalloc(sizeof(NetUnit_host));
	s->ip=ip;
	s->next=NULL;

	if (root==NULL) root=s;
	else last->next=s;
	last=s; 
	references++;
	pthread_mutex_unlock(lock);
	}

void NetUnit_cluster::Remove(struct in_addr ip){
	pthread_mutex_lock(lock);
	NetUnit_host *d, *p;
	for(d=root; d!=NULL; d=d->next)	{
		if (d->ip.s_addr==ip.s_addr) {
			if (d==root && d==last ) root=last=NULL;
			else if (d==root) root=d->next;
			else if (d==last) { last=p; last->next=NULL; }
			else p->next=d->next;

			references--;
			aFree(d);
			break;
			}
		p=d; 
	 	}
	pthread_mutex_unlock(lock);
	}

match NetUnit_cluster::Check(struct ip *ipc) {
	unsigned rs=0, rd=0;
	NetUnit_host *d;
	pthread_mutex_lock(lock);
	for (d=root; d!=NULL; d=d->next) {
		if (ntohl(d->ip.s_addr)==ntohl(ipc->ip_src.s_addr)) rs++;
		if (ntohl(d->ip.s_addr)==ntohl(ipc->ip_dst.s_addr)) rd++;
		}
	pthread_mutex_unlock(lock);
	if (rs && rd) return MATCH_BOTH;
	else if (rs) return MATCH_SRC;
	else if (rd) return MATCH_DST;
	else return MATCH_NONE;
	}

void NetUnit_cluster::unit2tree(IPTree *iptree,unsigned flag) {
	for(NetUnit_host *h=root;h!=NULL;h=h->next)
                                iptree->addr2tree(h->ip.s_addr,32,(NetUnit*)this,flag);
}
/////////////////////////////////////////////////////////////////////////
/////////////////////////////////////////////////////////////////////////
// NetUnit_group class
NetUnit_group::NetUnit_group() : NetUnit(NETUNIT_GROUP){
	references=0;
	}

match NetUnit_group::Check(struct ip *ipc) {
	NetUnit *t;
	unsigned rs=0, rd=0;
	match m;
	
	for (t=Units.root; t!=NULL; t=t->next)	
		if (t->parent==this) {
			m=t->Check(ipc);
			if (m==MATCH_BOTH) { rs++; rd++; }
			else if (m==MATCH_SRC) rs++;
			else if (m==MATCH_DST) rd++;
			}
				
	if (rs && rd) return MATCH_BOTH;
	else if (rs) return MATCH_SRC;
	else if (rd) return MATCH_DST;
	else return MATCH_NONE;
	}

/////////////////////////////////////////////////////////////////////////
// NetUnit_host class
NetUnit_host::NetUnit_host() : NetUnit(NETUNIT_HOST) {
	ip.s_addr=0;
	next=NULL;
	}

match NetUnit_host::Check(struct ip *ipc) {
	unsigned rs=0, rd=0;
	if (ntohl(ip.s_addr)==ntohl(ipc->ip_src.s_addr)) rs++;
	if (ntohl(ip.s_addr)==ntohl(ipc->ip_dst.s_addr)) rd++;

	if (rs && rd) return MATCH_BOTH;
	else if (rs) return MATCH_SRC;
	else if (rd) return MATCH_DST;
	else return MATCH_NONE;
	}

void NetUnit_host::unit2tree(IPTree *iptree,unsigned flag) {
	iptree->addr2tree(ip.s_addr,32,(NetUnit*)this,flag);
}
/////////////////////////////////////////////////////////////////////////
NetUnit_net::NetUnit_net() : NetUnit(NETUNIT_NET) {
	ip.s_addr=0;
	mask.s_addr=INADDR_BROADCAST;
	}

match NetUnit_net::Check(struct ip *ipc){
	unsigned rs=0, rd=0;
	
	//aDebug(DEBUG_DS_IP, "netunit_net chk <%s>: src:%lu, dst:%lu, net1:%lu, net2:%lu\n", name, ntohl(ipc->ip_src.s_addr), ntohl(ipc->ip_dst.s_addr), ntohl(ip.s_addr), ntohl(ip.s_addr)+INADDR_BROADCAST-ntohl(mask.s_addr) );

	if (ntohl(ip.s_addr)<ntohl(ipc->ip_src.s_addr) && ntohl(ipc->ip_src.s_addr)<(ntohl(ip.s_addr)+INADDR_BROADCAST-ntohl(mask.s_addr))) rs++;
	if (ntohl(ip.s_addr)<ntohl(ipc->ip_dst.s_addr) && ntohl(ipc->ip_dst.s_addr)<(ntohl(ip.s_addr)+INADDR_BROADCAST-ntohl(mask.s_addr))) rd++;

	if (rs && rd) return MATCH_BOTH;
	else if (rs) return MATCH_SRC;
	else if (rd) return MATCH_DST;
	else return MATCH_NONE;
	}

void NetUnit_net::unit2tree(IPTree *iptree,unsigned flag) {
	unsigned m=0;
	unsigned long tmp=mask.s_addr;
	
	for(unsigned i=0;i<32;i++) {
        	if(tmp&1) m++;
		tmp=tmp>>1;
	}
        iptree->addr2tree(ip.s_addr,m,(NetUnit*)this,flag);
}
/////////////////////////////////////////////////////////////////////////
// NetUnit_user class
NetUnit_user::NetUnit_user() : NetUnit(NETUNIT_USER) {
        ip.s_addr=0;
        next=NULL;
	real_name=email=password=NULL;
        }

match NetUnit_user::Check(struct ip *ipc) {
        unsigned rs=0, rd=0;
	if(ip.s_addr==0) return MATCH_NONE;
        if (ntohl(ip.s_addr)==ntohl(ipc->ip_src.s_addr)) rs++;
        if (ntohl(ip.s_addr)==ntohl(ipc->ip_dst.s_addr)) rd++;

        if (rs && rd) return MATCH_BOTH;
        else if (rs) return MATCH_SRC;
        else if (rd) return MATCH_DST;
        else return MATCH_NONE;
        }

void NetUnit_user::unit2tree(IPTree *iptree,unsigned flag) {
	iptree->addr2tree(ip.s_addr,32,(NetUnit*)this,flag);
}
/////////////////////////////////////////////////////////////////////////
unsigned cUnit(Connection *conn, char *param[], int no_flag){
	NetUnit *ut;
	int i=2;
	if (!strcmp(param[2], "name")) ut=Units.getUnit(param[3]);
	else if (!strcmp(param[2], "oid")) ut=Units.getUnitById(strtol(param[3], NULL, 16));
	else { aParse(conn, "unit %s name/oid unspecified\n", param[1]); return PARSE_OK; }

//-----------------------------------------------------------------------
	if (!strcasecmp(param[1], "host")) {
		NetUnit_host *u;
		u = (NetUnit_host*)ut;
		if (!ut && !no_flag) { 
			u = (NetUnit_host*)aMalloc(sizeof(NetUnit_host)); 
			aParse(conn, "unit host %06X created\n", u->id);
			NetUnit_host *t= new NetUnit_host(); 
			memcpy(u, t, sizeof(NetUnit_host));
			Units.Insert(u); 
			}
		if (!ut && no_flag) {
			aParse(conn, "unit host does not exist\n");
			return PARSE_OK;
			}
		if (ut && no_flag){
			aParse(conn, "unit host %06X deleted\n", u->id);
			Units.Delete(ut);
			ut->~NetUnit();
			return PARSE_OK;
			}
		ut=(NetUnit*)u;
		while (param[i]!=empty) {
	 			if (strcasecmp(param[i], "no")==0) { 
					no_flag=1; i-=1;

	 			} else if (strcasecmp(param[i], "name")==0) { 
					u->setName(param[i+1]);
					aParse(conn, "host %06X name set: %s\n", u->id, u->name);

				} else if (strcasecmp(param[i], "ip")==0) { 
					ut->unit2trees(0);
					inet_aton(param[i+1], &u->ip);
					ut->unit2trees(1);
					aParse(conn, "host %06X ip set: %s\n", u->id, inet_ntoa(u->ip));

				} else if (strcasecmp(param[i], "oid")==0) { 
					u->id=strtol(param[i+1], NULL, 16);
					aParse(conn, "host %06X oid set\n", u->id);

				} else if (strstr(param[i], "sys-")==param[i]) { // possibly sys-* here
					SetSysPolicy(param[i], u, conn);
					i-=1;

				} else if (strcasecmp(param[i], "acct-policy")==0) { 
					PolicyAdd(u, &i, POLICY_ACCT, conn, param, no_flag);

				} else if (strcasecmp(param[i], "fw-policy")==0) { 
					PolicyAdd(u, &i, POLICY_FW, conn, param, no_flag);
	
				} else if (strcasecmp(param[i], "ds-list")==0) { 
					ut->unit2trees(0);
					DSListAdd(u, &i, conn, param, no_flag);
					ut->unit2trees(1);

				} else if (strcasecmp(param[i], "parent")==0) {
					NetUnit_group *gr;
					if (param[i+1]==empty || no_flag) { 
						u->parent=NULL;
						aParse(conn, "host %06X parent cleared\n", u->id);
						}
					else {
						gr=(NetUnit_group*)Units.getUnit(param[i+1]);
						if (gr) { 
						u->parent=gr; 
							aParse(conn, "host %06X added to group %s\n", u->id, gr->name);
							}
						else aParse(conn, "group %s for host %06X not exist\n", param[i+1], u->id);
						}

				} else if (strcasecmp(param[i], "no-local-pass")==0) { 
					if (no_flag) u->nlp=0; else u->nlp=1;
					aParse(conn, "host %06X no-local-pass flag set: %u\n", u->id, u->nlp);	
					i-=1;

				} else aParse(conn, "host %06X command '%s' unknown\n", u->id, param[i]); 

			i=i+2;
			}
		return PARSE_OK;
		}
//-----------------------------------------------------------------------
	else if (!strcasecmp(param[1], "net")) {
		NetUnit_net *u;
		u = (NetUnit_net*)ut;
		if (!ut && !no_flag) { 
			u = (NetUnit_net*)aMalloc(sizeof(NetUnit_net)); 
			aParse(conn, "unit net %06X created\n", u->id); 
			NetUnit_net *t= new NetUnit_net(); 
			memcpy(u, t, sizeof(NetUnit_net));
			Units.Insert(u); 
			}
		
		if (!ut && no_flag) {
			aParse(conn, "unit net does not exist\n");
			return PARSE_OK;
			}

		if (ut && no_flag){
			aParse(conn, "unit net %06X deleted\n", u->id);
			Units.Delete(ut);
			ut->~NetUnit();
			return PARSE_OK;
			}

		ut=(NetUnit*)u;
		while (param[i]!=empty) {
	 		if (strcasecmp(param[i], "no")==0) { 
				no_flag=1; i-=1;

	 			} else if (strcasecmp(param[i], "name")==0) { 
				u->setName(param[i+1]);
				aParse(conn, "net %06X name set: %s\n", u->id, u->name);

				} else if (strcasecmp(param[i], "ip")==0) { 
				ut->unit2trees(0);
				inet_aton(param[i+1], &u->ip);
				ut->unit2trees(1);
				aParse(conn, "net %06X ip set: %s\n", u->id, inet_ntoa(u->ip));

				} else if (strcasecmp(param[i], "mask")==0) { 
				ut->unit2trees(0);
				inet_aton(param[i+1], &u->mask);
				ut->unit2trees(1);
				aParse(conn, "net %06X mask set: %s\n", u->id, inet_ntoa(u->mask));

				} else if (strcasecmp(param[i], "acct-policy")==0) { 
					PolicyAdd(u, &i, POLICY_ACCT, conn, param, no_flag);

				} else if (strcasecmp(param[i], "fw-policy")==0) { 
					PolicyAdd(u, &i, POLICY_FW, conn, param, no_flag);

				} else if (strcasecmp(param[i], "ds-list")==0) { 
					ut->unit2trees(0);
					DSListAdd(u, &i, conn, param, no_flag);
					ut->unit2trees(1);

				} else if (strcasecmp(param[i], "oid")==0) { 
					u->id=strtol(param[i+1], NULL, 16);
					aParse(conn, "net %06X oid set\n", u->id);

				} else if (strstr(param[i], "sys-")==param[i]) { // possibly sys-* here
					SetSysPolicy(param[i], u, conn);
					i-=1;

				} else if (strcasecmp(param[i], "parent")==0) {
				NetUnit_group *gr;
				if (param[i+1]==empty || no_flag) { 
					u->parent=NULL;
					aParse(conn, "net %06X parent cleared\n", u->id);
					}
				else {
					gr=(NetUnit_group*)Units.getUnit(param[i+1]);
					if (gr) { 
						u->parent=gr; 
						aParse(conn, "net %06X added to group %s\n", u->id, gr->name);
						}
					else aParse(conn, "group %s for net %06X not exist\n", param[i+1], u->id);
					}

				} else if (strcasecmp(param[i], "no-local-pass")==0) { 
					if (no_flag) u->nlp=0; else u->nlp=1;
					aParse(conn, "host %06X no-local-pass flag set: %u\n", u->id, u->nlp);	
					i-=1;

				} else aParse(conn, "net %06X command unknown\n", u->id); 

			i=i+2;
			}
		return PARSE_OK;
		}
//-----------------------------------------------------------------------
	else if (!strcasecmp(param[1], "cluster")) {
		NetUnit_cluster *u;
		u=(NetUnit_cluster*)ut;
		if (!ut && !no_flag) { 
			u = (NetUnit_cluster*)aMalloc(sizeof(NetUnit_cluster)); 
			aParse(conn, "unit cluster %06X created\n", u->id); 
			NetUnit_cluster *t= new NetUnit_cluster(); 
			memcpy(u, t, sizeof(NetUnit_cluster));
			Units.Insert(u); 
			}
		
		if (!ut && no_flag) {
			aParse(conn, "unit cluster does not exist\n");
			return PARSE_OK;
			}

		if (ut && no_flag){
			aParse(conn, "unit cluster %06X deleted\n", u->id);
			Units.Delete(ut);
			ut->~NetUnit();
			return PARSE_OK;
			}
		
		ut=(NetUnit*)u;
		while (param[i]!=empty) {
	 		if (strcasecmp(param[i], "no")==0) { 
				no_flag=1; i-=1;

	 			} else if (strcasecmp(param[i], "name")==0) { 
				u->setName(param[i+1]);
				aParse(conn, "cluster %06X name set: %s\n", u->id, u->name);

				} else if (strcasecmp(param[i], "ip")==0) { 
				struct in_addr addr;
				ut->unit2trees(0);
				inet_aton(param[i+1], &addr);
				u->Add(addr);
				ut->unit2trees(1);
				aParse(conn, "cluster %06X ip added: %s\n", u->id, inet_ntoa(addr));

				} else if (strcasecmp(param[i], "no")==0 && strcasecmp(param[i+1], "ip")==0) { 
				struct in_addr addr;
				ut->unit2trees(0);
				inet_aton(param[i+2], &addr);
				u->Remove(addr);
				ut->unit2trees(1);
				aParse(conn, "cluster %06X ip removed: %s\n", u->id, inet_ntoa(addr));
				i++;

				} else if (strcasecmp(param[i], "acct-policy")==0) { 
					PolicyAdd(u, &i, POLICY_ACCT, conn, param, no_flag);

				} else if (strcasecmp(param[i], "fw-policy")==0) { 
					PolicyAdd(u, &i, POLICY_FW, conn, param, no_flag);

				} else if (strcasecmp(param[i], "ds-list")==0) { 
					ut->unit2trees(0);
					DSListAdd(u, &i, conn, param, no_flag);
					ut->unit2trees(1);

				} else if (strcasecmp(param[i], "oid")==0) { 
					u->id=strtol(param[i+1], NULL, 16);
					aParse(conn, "cluster %06X oid set\n", u->id);

				} else if (strstr(param[i], "sys-")==param[i]) { // possibly sys-* here
					SetSysPolicy(param[i], u, conn);
					i-=1;

				} else if (strcasecmp(param[i], "parent")==0) {
				NetUnit_group *gr;
				if (param[i+1]==empty || no_flag) { 
					u->parent=NULL;
					aParse(conn, "cluster %06X parent cleared\n", u->id);
					}
				else {
					gr=(NetUnit_group*)Units.getUnit(param[i+1]);
					if (gr && gr->name) { 
						u->parent=gr; 
						aParse(conn, "cluster %06X added to group %s\n", u->id, gr->name);
						}
					else aParse(conn, "group %s for cluster %06X not exist\n", param[i+1], u->id);
					}

				} else if (strcasecmp(param[i], "no-local-pass")==0) { 
					if (no_flag) u->nlp=0; else u->nlp=1;
					aParse(conn, "host %06X no-local-pass flag set: %u\n", u->id, u->nlp);	
					i-=1;

				} else aParse(conn, "cluster %06X command unknown\n", u->id); 

			i=i+2;
			}
		return PARSE_OK;
		}
//-----------------------------------------------------------------------
	else if (!strcasecmp(param[1], "group")) {
		NetUnit_group *u;
		u=(NetUnit_group*)ut;
		if (!ut && !no_flag) { 
			u = (NetUnit_group*)aMalloc(sizeof(NetUnit_group)); 
			aParse(conn, "unit group %06X created\n", u->id); 
			NetUnit_group *t= new NetUnit_group(); 
			memcpy(u, t, sizeof(NetUnit_group));
			Units.Insert(u); 
			}
		
		if (!ut && no_flag) {
			aParse(conn, "unit group does not exist\n");
			return PARSE_OK;
			}

		if (ut && no_flag){
			aParse(conn, "unit group %06X deleted\n", u->id);
			Units.Delete(ut);
			ut->~NetUnit();
			return PARSE_OK;
			}

		while (param[i]!=empty) {
			NetUnit_group *gr;
	 		if (strcasecmp(param[i], "no")==0) { 
				no_flag=1; i-=1;

	 			} else if (strcasecmp(param[i], "name")==0) { 
				u->setName(param[i+1]);
				aParse(conn, "group %06X name set: %s\n", u->id, u->name);

				} else if (strcasecmp(param[i], "parent")==0) { 
				gr=(NetUnit_group*)Units.getUnit(param[i+1]);
				if (gr && gr->name) {
					u->parent=gr;
					aParse(conn, "group %06X parent %s set\n", u->id, gr->name);
					} else aParse(conn, "group %s cannot find parent\n", param[i+1]);

				} else if (strcasecmp(param[i], "no")==0 && strcasecmp(param[i+1], "parent")==0) {
					u->parent=NULL;
					aParse(conn, "group %06X parent cleared\n", u->id);
					i++;

				} else if (strcasecmp(param[i], "acct-policy")==0) { 
					PolicyAdd(u, &i, POLICY_ACCT, conn, param, no_flag);

				} else if (strcasecmp(param[i], "fw-policy")==0) { 
					PolicyAdd(u, &i, POLICY_FW, conn, param, no_flag);

				} else if (strcasecmp(param[i], "ds-list")==0) { 
					DSListAdd(u, &i, conn, param, no_flag);

				} else if (strcasecmp(param[i], "oid")==0) { 
					u->id=strtol(param[i+1], NULL, 16);
					aParse(conn, "group %06X oid set\n", u->id);

				} else if (strstr(param[i], "sys-")==param[i]) { // possibly sys-* here
					SetSysPolicy(param[i], u, conn);
					i-=1;

				} else if (strcasecmp(param[i], "no-local-pass")==0) { 
					if (no_flag) u->nlp=0; else u->nlp=1;
					aParse(conn, "host %06X no-local-pass flag set: %u\n", u->id, u->nlp);	
					i-=1;

				} else aParse(conn, "group %06X command unknown\n", u->id); 

			i=i+2;
			}
		return PARSE_OK;
		}
//-----------------------------------------------------------------------
        if (!strcasecmp(param[1], "user")) {
                NetUnit_user *u;
                u = (NetUnit_user*)ut;
                if (!ut && !no_flag) {
                        u = (NetUnit_user*)aMalloc(sizeof(NetUnit_user));
                        aParse(conn, "unit user %06X created\n", u->id);
                        NetUnit_user *t= new NetUnit_user();
                        memcpy(u, t, sizeof(NetUnit_user));
                        Units.Insert(u);
                        }
                if (!ut && no_flag) {
                        aParse(conn, "unit user does not exist\n");
                        return PARSE_OK;
                        }
                if (ut && no_flag){
                        aParse(conn, "unit user %06X deleted\n", u->id);
                        Units.Delete(ut);
                        ut->~NetUnit();
                        return PARSE_OK;
                        }
		ut=(NetUnit*)u;

                while (param[i]!=empty) {
                                if (strcasecmp(param[i], "no")==0) {
                                        no_flag=1; i-=1;

                                } else if (strcasecmp(param[i], "name")==0) {
                                        u->setName(param[i+1]);
                                        aParse(conn, "user %06X name set: %s\n", u->id, u->name);
				
				} else if (strcasecmp(param[i], "password")==0) {
                                        aFree(u->password);
        				u->password=(char *)aMalloc(strlen(param[i+1]));
        				strcpy(u->password, param[i+1]);
                                        aParse(conn, "user %06X password set: %s\n", u->id, u->password);
				
				} else if (strcasecmp(param[i], "real_name")==0) {
                                        aFree(u->real_name);
                                        u->real_name=(char *)aMalloc(strlen(param[i+1]));
                                        strcpy(u->real_name, param[i+1]);
                                        aParse(conn, "user %06X real_name set: %s\n", u->id, u->real_name);
				
				} else if (strcasecmp(param[i], "email")==0) {
                                        aFree(u->email);
                                        u->email=(char *)aMalloc(strlen(param[i+1]));
                                        strcpy(u->email, param[i+1]);
                                        aParse(conn, "user %06X email set: %s\n", u->id, u->email);

                                } else if (strcasecmp(param[i], "ip")==0) {
                                        inet_aton(param[i+1], &u->ip);
                                        aParse(conn, "user %06X ip set: %s\n", u->id, inet_ntoa(u->ip));

                                } else if (strcasecmp(param[i], "oid")==0) {
					ut->unit2trees(0);
                                        u->id=strtol(param[i+1], NULL, 16);
					ut->unit2trees(1);
                                        aParse(conn, "user %06X oid set\n", u->id);

                                } else if (strstr(param[i], "sys-")==param[i]) { // possibly sys-* here
                                        SetSysPolicy(param[i], u, conn);
                                        i-=1;

                                } else if (strcasecmp(param[i], "acct-policy")==0) {
                                        PolicyAdd(u, &i, POLICY_ACCT, conn, param, no_flag);

                                } else if (strcasecmp(param[i], "fw-policy")==0) {
                                        PolicyAdd(u, &i, POLICY_FW, conn, param, no_flag);

                                } else if (strcasecmp(param[i], "ds-list")==0) {
					ut->unit2trees(0);
                                        DSListAdd(u, &i, conn, param, no_flag);
					ut->unit2trees(1);

                                } else if (strcasecmp(param[i], "parent")==0) {
                                        NetUnit_group *gr;
                                        if (param[i+1]==empty || no_flag) {
                                                u->parent=NULL;
                                                aParse(conn, "user %06X parent cleared\n", u->id);
                                                }
                                        else {
                                                gr=(NetUnit_group*)Units.getUnit(param[i+1]);
                                                if (gr) {
                                                u->parent=gr;
                                                        aParse(conn, "user %06X added to group %s\n", u->id, gr->name);
                                                        }
                                                else aParse(conn, "group %s for user %06X not exist\n", param[i+1], u->id);
                                                }

                                } else if (strcasecmp(param[i], "no-local-pass")==0) {
                                        if (no_flag) u->nlp=0; else u->nlp=1;
                                        aParse(conn, "user %06X no-local-pass flag set: %u\n", u->id, u->nlp);
                                        i-=1;

                                } else aParse(conn, "user %06X command '%s' unknown\n", u->id, param[i]);

                        i=i+2;
                        }
                return PARSE_OK;
                }
//-----------------------------------------------------------------------
	else aParse(conn, "unit type unknown\n");

	return PARSE_OK;
	}

/////////////////////////////////////////////////////////////////////////
unsigned cShowUnits(Connection *conn){	   // "show units"
	pthread_rwlock_rdlock(Units.rwlock);
	NetUnit *d;
	char *parent_name;

	fprintf(conn->stream_w, "%8s | %6s | %10s | %3s | %10s | %-20s\n", "TYPE", "OID", "NAME", "NLP", "PARENT", "PARAMS");
	for(d=Units.root; d!=NULL; d=d->next)	switch (d->type) {
		case NETUNIT_HOST: {
			NetUnit_host *h;
			if (d->parent && d->parent->name) parent_name=d->parent->name; else parent_name="<>";
			h = (NetUnit_host*)d;
			fprintf(conn->stream_w, "%8s | %06X | %10s | %3s | %10s | IP: %-12s\n", "host", h->id, h->name?h->name:"<??>", h->nlp?" + ":"", parent_name, inet_ntoa(h->ip));
			} break;
		case NETUNIT_USER: {
			NetUnit_user *h;
			if (d->parent && d->parent->name) parent_name=d->parent->name; else parent_name="<>";
			h = (NetUnit_user*)d;
			fprintf(conn->stream_w, "%8s | %06X | %10s | %3s | %10s | IP: %-12s\n", "user", h->id, h->name?h->name:"<??>", h->nlp?" + ":"", parent_name, inet_ntoa(h->ip));
			} break;
		case NETUNIT_CLUSTER: { 
			NetUnit_cluster *h;
			NetUnit_host *t;
			if (d->parent && d->parent->name) parent_name=d->parent->name; else parent_name="<>";
			h = (NetUnit_cluster*)d;
			fprintf(conn->stream_w, "%8s | %06X | %10s | %3s | %10s | IPs: ", "cluster", h->id, h->name?h->name:"<??>", h->nlp?" + ":"", parent_name);
			for (t=h->root; t!=NULL; t=t->next)	fprintf(conn->stream_w, "%s ", inet_ntoa(t->ip));
			fprintf(conn->stream_w, "\n");
			} break;
		case NETUNIT_GROUP: {
			NetUnit_group *h;
			NetUnit *t; 
			if (d->parent && d->parent->name) parent_name=d->parent->name; else parent_name="<>";
			h = (NetUnit_group*)d;
			fprintf(conn->stream_w, "%8s | %06X | %10s | %3s | %10s | Sub: ", "group", h->id, h->name?h->name:"<??>", h->nlp?" + ":"", parent_name);
			for (t=Units.root; t!=NULL; t=t->next)	if (t->parent==d && t->name) fprintf(conn->stream_w, "%s ", t->name);
			fprintf(conn->stream_w, "\n");
			} break;
		case NETUNIT_NET: {
			NetUnit_net *h;
			if (d->parent && d->parent->name) parent_name=d->parent->name; else parent_name="<>";
			h = (NetUnit_net*)d;
			fprintf(conn->stream_w, "%8s | %06X | %10s | %3s | %10s | %s", "net", h->id, h->name?h->name:"<??>", h->nlp?" + ":"", parent_name, inet_ntoa(h->ip));
			fprintf(conn->stream_w, ":%s\n", inet_ntoa(h->mask));
			} break;
		default: 
			fprintf(conn->stream_w, "%8s | %06X\n", "<??>", d->id);
			break;
		}

	pthread_rwlock_unlock(Units.rwlock);
	return PARSE_OK;
	}

/////////////////////////////////////////////////////////////////////////
/////////////////////////////////////////////////////////////////////////
unsigned cShowUnitsList(Connection *conn, int isfull, char *p1, char *p2){ // "show list {full}"
	pthread_rwlock_rdlock(Units.rwlock);
	NetUnit *d, *d2=NULL;
	policy_data *cpd;
	char *parent_name, *t_str;
	t_str=(char *)aMalloc(32);

	if (p1!=empty && p2!=empty) {
		if (!strcasecmp(p1, "name")) d2=Units.getUnit(p2);
		else if (!strcasecmp(p1, "oid")) d2=Units.getUnitById(strtol(p2, NULL, 16));
		}

	for (d=Units.root; d!=NULL; d=d->next) {
  		
  		if (p1!=empty && p2!=empty && d!=d2) continue;
  		
  		fprintf(conn->stream_w, "OID: %06X Name: %-10s Type: ", d->id, d->name);
  		switch (d->type){
			case NETUNIT_HOST: fprintf(conn->stream_w, "%-8s", "host"); break;  
			case NETUNIT_NET: fprintf(conn->stream_w, "%-8s", "net"); break;  
			case NETUNIT_CLUSTER: fprintf(conn->stream_w, "%-8s", "cluster"); break;  
			case NETUNIT_GROUP: fprintf(conn->stream_w, "%-8s", "group"); break; 
			case NETUNIT_USER: fprintf(conn->stream_w, "%-8s", "user"); break;
			default: fprintf(conn->stream_w, "%-8s", "<??>"); break;  
  			}
		if (d->nlp) fprintf(conn->stream_w, " NLP"); 
		if (d->parent && d->parent->name) parent_name=d->parent->name; else parent_name="<>";
		fprintf(conn->stream_w, " Parent: %-10s", parent_name);

		if (d->dsl_root) {
			fprintf(conn->stream_w, " DS-list: ");
			for (DSList *dsl=d->dsl_root;dsl!=NULL;dsl=dsl->next)
				 fprintf(conn->stream_w, "%u ", dsl->id);
		}

		fprintf(conn->stream_w, "\n");
				
		fprintf(conn->stream_w, " %-11s", "SYST policy");
			if (d->sys_policy==SP_NONE) fprintf(conn->stream_w, " is not set\n");
			else {
				char *buf=NULL; 
				buf=ShowSysPolicy(buf, d->sys_policy,d->sys_policy_perm, d->sys_policy_name);    
				fprintf(conn->stream_w, ": %s\n", buf);
				aFree(buf);
				} 

		fprintf(conn->stream_w, " %-11s", "  FW policy");
		if (d->fp.root)	fprintf(conn->stream_w, ": %-6s %-10s %-12s %-12s\n",  "OID", "NAME", "CHECK", "MATCH");
		else fprintf(conn->stream_w, " list is empty\n");

		for (cpd=d->fp.root; cpd!=NULL; cpd=cpd->next) {
			fprintf(conn->stream_w, "%-10s %s%s %06X %-10s %-12qu %-12qu\n", "", cpd->brk?"%":" ", cpd->inv?"!":" ", cpd->policy->id,  cpd->policy->name?cpd->policy->name:"<??>", cpd->check, cpd->match);
//			t_str=timeU2T(cpd->flow.from, NULL);
//			if (isfull) fprintf(conn->stream_w, " %-19s %-10s in: %-12qu out: %-12qu\n", t_str,  "flow", cpd->flow.in, cpd->flow.out);
//			aFree(t_str);
			}
						  
		fprintf(conn->stream_w, " %-11s", "ACCT policy");
		if (d->ap.root)	fprintf(conn->stream_w, ": %-6s %-10s %-12s %-12s\n",  "OID", "NAME", "CHECK", "MATCH");
		else fprintf(conn->stream_w, " list is empty\n");

		for (cpd=d->ap.root; cpd!=NULL; cpd=cpd->next) {
			fprintf(conn->stream_w, "%-11s %s %06X %-10s %-12qu %-12qu\n", "", cpd->inv?"!":" ", cpd->policy->id, cpd->policy->name?cpd->policy->name:"<??>", cpd->check, cpd->match);
			if (isfull) {
				timeU2T(cpd->flow.from, t_str); fprintf(conn->stream_w, " %-19s %-10s in: %-12qu out: %-12qu\n", cpd->flow.from?t_str:"--.--.---- --:--:--",  "flow", cpd->flow.in, cpd->flow.out);
				timeU2T(cpd->t.from, t_str); fprintf(conn->stream_w, " %-19s %-10s in: %-12qu out: %-12qu\n", t_str,  "total", cpd->t.in, cpd->t.out);
				timeU2T(cpd->m.from, t_str); fprintf(conn->stream_w, " %-19s %-10s in: %-12qu out: %-12qu\n", t_str,  "month", cpd->m.in, cpd->m.out);
				timeU2T(cpd->w.from, t_str); fprintf(conn->stream_w, " %-19s %-10s in: %-12qu out: %-12qu\n", t_str,  "week", cpd->w.in, cpd->w.out);
				timeU2T(cpd->d.from, t_str); fprintf(conn->stream_w, " %-19s %-10s in: %-12qu out: %-12qu\n", t_str,  "day", cpd->d.in, cpd->d.out);
				timeU2T(cpd->h.from, t_str); fprintf(conn->stream_w, " %-19s %-10s in: %-12qu out: %-12qu\n", t_str,  "hour", cpd->h.in, cpd->h.out);
				}			
			}

		fprintf(conn->stream_w, "\n");
		}
	aFree(t_str);
	pthread_rwlock_unlock(Units.rwlock);
	return PARSE_OK;
	}

/////////////////////////////////////////////////////////////////////////
void DSListAdd(NetUnit *u, int *i, Connection *conn, char *param[32], int no_flag){
	char *res, *ptr;

	aDebug(DEBUG_DS_IP, "NetUnit: DSL: adding DS list \"%s\"\n", param[(*i)+1]);

	int k=strlen(param[(*i)+1]);
	unsigned f;

	for (ptr=param[(*i)+1]; (ptr-param[(*i)+1])<k ;ptr=res+1){
		if(no_flag && ( !strcmp(ptr,"all") || !strcmp(ptr,"*")) ) {
			aParse(conn, "ds-list cleared for unit %06X\n", u->id);
			DSList *ptr,*dsl=u->dsl_root;
			while(dsl) {
				ptr=dsl;
				dsl=dsl->next;
				aFree(ptr);
			}
			u->dsl_root=NULL;
			return;
		}
		f=strtol(ptr, &res, 10);
		if (no_flag) {
			aParse(conn, "removing DS %u from list for unit %06X\n", f, u->id);
			u->clearDSList(f);
		}
		else {
			aParse(conn, "adding DS %u to list for unit %06X\n", f, u->id);
			u->setDSList(f);
		}
	}
}

/////////////////////////////////////////////////////////////////////////
